//
// Avi2Flc by ssjx (http://ssjx.co.uk)
//
// Compile:
//		dmd avi2flc.d fli.d msvc.d
// or
//		dmd -i avi2flc.d	
//
//
// 0.5 Changes (20/05/2023)
//  20/05/2023 - Ring frame creation fixed again (old code not removed)...
//  20/05/2023 - Disabled best mode as not working correctly
//
//
// 0.4 Changes
// 	21/02/2023 - Better AVI parser
//  22/02/2023 - Incorrect/unnecessary size checks
//  22/02/2023 - Ring frame creation improved
//  23/02/2023 - Changed handler check/case change
//
// 0.3 Changes
//	+ Makes use of modules so easier to compile and better contained
//  + MSVC decoder! (Best mode does not work with msvc avi files...)
//  + Immutables
//  + Some cast() removed due to more accurate shifts
//
// 0.2 Changes
//	+ Some loops changed to use foreach()
// 	+ Can now handle 32bit AVI files 
// 	+ Added switch '-best' for better colours, slow though...
// 	+ RLE compression!

import std.stdio;
import std.file;
import std.math; //abs()

import fli,msvc;

// 56 
struct Avih{	
	align(1):
	int microps;
	int maxbps;
	int res;
	int flags;
	int totf;
	int initf;
	int streams;
	int buffersize;
	int biWidth;
	int biHeight;
	int scale;
	int fps;
	int strt;
	int leng;
};

//
struct Vsdhr{	
	align(1):
	char[4] subtype;
	char[4] handler; // e.g. MSVC
	int flags;
	int pri;
	int init;
	int scale;
	int fps;
	int strt;
	int lng;
	int bsize;
	int quality;
	int ssize;
	short[4] rectframe;
};

// 40
struct Vstrf{	
	align(1):
	int biSize;
	int biWidth;
	int biHeight;
	short biPlanes;
	short biBitCount;
	int biCompression;
	int biSizeImage;
	int biXPelsPerMeter;
	int biYPelsPerMeter;
	int biClrUsed;
	int biClrImportant;
};

// 8 
struct Hdr{	
	align(1):
	char[4] db;
	int size;
};

// ---
Avih[1] avi_header;
Vsdhr[1] str_header;
Vstrf[1] str_format;
Hdr[1] header;	
// ---
struct Newchunk{	
	align(1):
	int size;
	char[4] chunktype;
};

Newchunk[1] nchunk;
//


Myflihdr[1] fli_header;
Myfrmhdr[1] frm_header;
Mychkhdr[1] chk_header;

Mypal[1] pal_header;
Myrgb[256] fli_rgb;

//
// Fixed 256 palette
//

@safe void palette_set()
{
	int c=0;
	
	for(ubyte r=0;r<8;r++){
		for(ubyte g=0;g<8;g++){
			for(ubyte b=0;b<4;b++){
				fli_rgb[c].r=cast(ubyte)(r*36);
				fli_rgb[c].g=cast(ubyte)(g*36);
				fli_rgb[c].b=cast(ubyte)(b*85);
				c++;
			}
		}
	}
}


//
// Convert 24/32bit bitmap to fixed palette 8bit
//
ubyte[] bmp;
ubyte[] frame;	// FLI 8bit frame

@safe void convert(const int w, const int h,const int d)
{
	int pos=0;

	immutable int bpp=(d>>3);
	
	foreach(y;0..h)
	{
		int c=((h-1)-y)*(w*bpp);	// flip image
	
		foreach(x;0..w)
		{	
			ubyte b=(bmp[c] & 0xff)>>6;
			ubyte g=(bmp[c+1] & 0xff)>>5;
			ubyte r=(bmp[c+2] & 0xff)>>5;
			
			// Being exact with shift meant I could remove cast()
			ubyte ar=(r & 0x07)<<5;
			ubyte ag=(g & 0x1f)<<2;
			ubyte ab=b;
			
			frame[pos]=cast(ubyte)(ar+ag+ab);
		
			c+=bpp;
			pos++;
		}	
	}
}


//
// Better Optimised 256 Palette 
//

int totalcol=0;
int nextcol=1;

@safe void resetpal(){
	totalcol=0;
	nextcol=1;
	
	foreach(i;0..256){
		fli_rgb[i].r=0;
		fli_rgb[i].g=0;
		fli_rgb[i].b=0;
	}
}

@safe ubyte getcol(const ubyte r,const ubyte g,const ubyte b){
	ubyte idx;
	int err,lowest=999;
	
	foreach(ubyte i;0..256){
		err=abs(fli_rgb[i].r-r);
		err+=abs(fli_rgb[i].g-g);
		err+=abs(fli_rgb[i].b-b);
		
		if (err<lowest){
			lowest=err;
			idx=i;
		}
	}
	
	return idx;
}

//
// Convert 24/32bit bitmap to best palette 8bit
//

@safe void convert2(const int w, const int h,const int d){
	ubyte r,g,b;

	int c=0;
	int pos=0;
	
	immutable int bpp=(d>>3);
	
	foreach(y;0..h){
		c=((h-1)-y)*(w*bpp);	// flip image
	
		foreach(x;0..w){	
			b=bmp[c];
			g=bmp[c+1];
			r=bmp[c+2];
			
			frame[pos]=getcol(r,g,b);
		
			c+=bpp;
			pos++;
		}	
	}
}


@safe bool inpal(const ubyte r,const ubyte g,const ubyte b,const int e){
	int err=0;
	
	foreach(i;0..256){
		err=abs(fli_rgb[i].r-r);
		err+=abs(fli_rgb[i].g-g);
		err+=abs(fli_rgb[i].b-b);
		if (err<=e){return true;}
	}
		
	if (nextcol<=255){
		fli_rgb[nextcol].r=r;
		fli_rgb[nextcol].g=g;
		fli_rgb[nextcol].b=b;
		nextcol++;
		return true;
	}
	
	// Restart!
	return false;	
} 

@safe void optpalette(int width, int height,int depth)
{
	int err;
	bool ok=true;
	immutable int bpp=(depth>>3);
	do{
		
		resetpal();
		int c=0;
		
	    foreach(y;0..height){ 	
		    foreach(x;0..width){
	    		ok=inpal(bmp[c+2],bmp[c+1],bmp[c],err );		
	    		if (ok==false){break;}
	    		c=c+bpp;
		    }
			if (ok==false){break;}
        }
        
    	if (ok==false){err+=10;}	// changing this to a smaller number reduces palette error but slow
	}while(ok!=true);

//	writeln("Palette Error:",err," Colours:",totalcol);
}

//
//
//

void main(string[] args)
{
//	writeln("Avi Header: ",aheader.sizeof);
//	writeln("V Header: ",vheader.sizeof);
//	writeln("V Format: ",vformat.sizeof);
//	writeln("Chunk: ",chunk.sizeof);
//	writeln("Header: ",header.sizeof);

	bool best_colour=false;
	bool option_debug=false;
	string fn;
	
	writeln("Avi2Flc v0.5 D (20/05/2023) by ssjx (http://ssjx.co.uk)");
	writeln("-------------------------------------------------------");
	
	if (args.length>1){
		foreach(par;args){
			if (par[0]=='-'){
				// if (par=="-best"){best_colour=true;}
				if (par=="-debug"){option_debug=true;}
			}else{
				if (par!=args[0]){
					// Not the exe or an option so probably an avi file
					fn=par;
				}
			}
		}
	}else{
		writeln("Usage: avi2flc [option] myfile.avi\n");
		writeln("Options:");
		// writeln("\t-best  : Optimised palette per frame. Slow!");
		writeln("\t-debug : Lists frame method and sizes");
		writeln("\nAvi should be uncompressed or MSVC encoded and the width divisble by 4.");
		return;
	}
	

	// Start reading AVI file
	if (exists(fn))
	{	
		string outfn=fn~".flc";	// <-- Note squiggle used to join strings!
	
		writeln("In: ",fn);
		writeln("Out: ",outfn);
	
		try
		{
			bool ismovie=false;	
			File file=File(fn, "rb");
	    
	    	file.seek(12);
	    
			// New AVI parser
			char[4] avi_list;
			char[4] subtype;
			// char[4] handler;
			
			while(true) {
				int size;
				file.rawRead(avi_list);
				
				char[4] ctype=avi_list;

				if (ctype=="LIST") {
					file.rawRead(nchunk);
					size = nchunk[0].size;
					ctype=nchunk[0].chunktype;
					writeln("List : ",ctype);

					switch (ctype){
					case "movi":
						writeln("Movie List!");
						ismovie = true;
					break;
					case "hdrl":
						writeln("Header List");
					break;
					case "strl":
						 writeln("Stream List");
					break;
					default:
						writeln("Unknown List ",ctype);
					break;
					}
				} else {
					// Chunk
					int[1] avi_chunksize;
					file.rawRead(avi_chunksize);
					size = avi_chunksize[0];	

					switch (ctype){
					case "avih":
						// AVI file header
						writeln("+ Avi Header ",size);
						file.rawRead(avi_header);
						int bmhd_frames=avi_header[0].totf;
						writeln("  Frames: ", bmhd_frames);
					break;
					
					case "strh":
						// Stream header
						writeln("+ Stream Header ",size);
						file.rawRead(str_header);
						// subtype = to!string(str_header[0].subtype);
						subtype = str_header[0].subtype;

						if (subtype=="vids"){
							
							// Lower to upper (if needed)
							foreach(ref h;str_header[0].handler){
								if (h>=97 && h<=122){h-=32;}
							}
							
							writeln("  Format: [",str_header[0].handler,"]");

							int scale=str_header[0].scale;
							int fps=str_header[0].fps;
							int bmhd_fps = fps / scale;

							 writeln("  Rate: ", fps, " / ", scale, " [", bmhd_fps, "]");

						} else {
							 writeln("  Ignoring: ",subtype);
						}
					break;
					
					case "strf":
						// Stream format
						writeln("+ Stream Format ",size);
						file.rawRead(str_format);

						if (subtype=="vids"){
							int bmhd_w = str_format[0].biWidth;
							int bmhd_h = str_format[0].biHeight;
							int bmhd_d = str_format[0].biBitCount;
							int bmhd_bytes = bmhd_d / 8;

							 writeln("  Size: ", bmhd_w, " x ", bmhd_h, " x ", bmhd_d);
						} else {
							 writeln("  Ignoring: ",subtype);
						}
					break;
					
					case "JUNK":
						writeln("+ Junk ",size);
						file.seek(size,SEEK_CUR);	
					break;
					
					default:
						writeln("+ Unknown ", ctype, " Size:", size);
						file.seek(size,SEEK_CUR);
					break;
					}
				}

				if (size==0){
					 writeln("Panic Exit!");
					ismovie = false;
					break;
				}

				if (ismovie==true){
					break;
				}
			}
							
			//
			if ((str_header[0].handler!= "DIB ") && (str_header[0].handler!="MSVC")) {
				writeln("Unsupported compression [", str_header[0].handler, "], video must be uncompressed DIB or MSVC.");
				return;
			}
		
			if ((str_format[0].biWidth % 4) > 0) {
				writeln("AVI width must be divisible by 4");
				return;
			}
			
			//

			if (ismovie==true)
			{		
				try
				{
					// Uncompressed frame so all are the same size
					int fsize=str_format[0].biWidth*str_format[0].biHeight;
					frame.length=fsize;
					
					if (str_header[0].handler=="MSVC"){
						if (best_colour==true){writeln("Best colour mode only works with uncompressed 24bit AVI files");}
						best_colour=false;
					}
					
					// Lazy 3:3:2 palette
					if (best_colour==false){
						palette_set();
					}
					
					pal_header[0].packets=1;
					pal_header[0].skip=0;
					pal_header[0].change=0;
				
					// Frame header (changes after first frame then stays same)
					frm_header[0].size=16+6+4+fli_rgb.sizeof+6+(fsize);
					frm_header[0].magic=cast(short)0xf1fa;
					frm_header[0].chunks=2;	// Palette + Frame
					
					//
					immutable short fps=(str_header[0].scale*1000)/str_header[0].fps;
					immutable short frames=cast(short)avi_header[0].totf;
					
					fli_header[0].magic=cast(short)0xAF12;
					fli_header[0].width=cast(short)str_format[0].biWidth;
					fli_header[0].height=cast(short)str_format[0].biHeight;
					fli_header[0].depth=8;
					fli_header[0].speed=fps;//1000/fps;
					fli_header[0].frames=frames;
					
					// +1 to include ring
					fli_header[0].size=128+(frm_header[0].size*(fli_header[0].frames+1));
					
					/*
					writeln("File Size: ",fli_header[0].size);
					writeln("Frm hdr Size: ",frm_header[0].size);
					writeln("FLI header: ",fli_header.sizeof);
					writeln("FLI Frm header: ",frm_header.sizeof);
					writeln("Palette Size: ",fli_rgb.sizeof);
					*/
		
					File outfile = File(outfn, "wb");   	
			        writeln("Saving as ",outfn);	
					outfile.rawWrite(fli_header);
		
					ubyte[] data;	// (Un)Compressed frame FLC data to write to file
					ubyte[] first;
					
					ubyte frame_type;
					ubyte[] avi_frame;
					
					bmp.length=str_format[0].biWidth*str_format[0].biHeight*3;
					
					foreach(f;0..frames+1)
					{
						// Reset each loop as we do not know the compressed frames size
						data.length=str_format[0].biWidth*str_format[0].biHeight;	// This should shrink later...
					
						if (f!=frames){
							// Read a frame from the AVI
							file.rawRead(header);
							
							// This will change each frame if not an RGB/DIB avi...
							avi_frame.length=header[0].size;
							file.rawRead(avi_frame);
							
							// Make a copy of first AVI frame to make ring frame from
							if (f==0){
								first.length=avi_frame.length;
								first[]=avi_frame[];
							}
						
							//
						}else{
							// Ring is a copy of the first frame
							if (option_debug==true){writeln("Ring/Loop Frame:");}
							avi_frame.length=first.length;
							avi_frame[]=first[];
						}
						
						//
						
						final switch(str_header[0].handler){
							case "DIB ":
								bmp[]=avi_frame[];
								
								if (best_colour==true){
									optpalette(str_format[0].biWidth,str_format[0].biHeight,str_format[0].biBitCount);
									convert2(str_format[0].biWidth,str_format[0].biHeight,str_format[0].biBitCount);
								}else{
									// 24bit > 8bit
									convert(str_format[0].biWidth,str_format[0].biHeight,str_format[0].biBitCount);	
								}
							break;
							
							case "MSVC":
								// Decompresses and converts 24bit > 8bit
								decode_msvc(avi_frame, frame, str_format[0].biWidth, str_format[0].biHeight);
							break;
							}
						
						// Sort out the colours...
						
						
						if (f>0){
							fsize=encode_rle(frame,data,str_format[0].biWidth,str_format[0].biHeight);
							frame_type=15;	// FLI Brun (RLE)
					
							// If encode_rle does not work, size returned is -1
							if (fsize<0){
								encode_copy(frame,data);
								fsize=str_format[0].biWidth*str_format[0].biHeight;
								frame_type=16;	// FLI Copy
							}
						}else{
							// First frame is always a FLI Copy
							encode_copy(frame,data);
							fsize=str_format[0].biWidth*str_format[0].biHeight;
							frame_type=16;		// FLI Copy
						}
						
						data.length=fsize;
						if (option_debug==true){
							if (frame_type==15){writeln(f,". RLE  ",fsize);}
							if (frame_type==16){writeln(f,". Copy ",fsize);}
						}
						
						if (best_colour==true || f==0){
							//
							frm_header[0].size=16+6+4+fli_rgb.sizeof+6+(fsize);
							frm_header[0].magic=cast(short)0xf1fa;
							frm_header[0].chunks=2;	// Image and Palette data	
							outfile.rawWrite(frm_header);
						
							// Write Palette
							chk_header[0].size=6+4+fli_rgb.sizeof;
							chk_header[0].typ=4; // FLI 256 Palette
							outfile.rawWrite(chk_header);
							outfile.rawWrite(pal_header);
							outfile.rawWrite(fli_rgb);	
						}else{
							frm_header[0].size=16+6+(fsize);
							frm_header[0].magic=cast(short)0xf1fa;
							frm_header[0].chunks=1;	// Image data only
							
							outfile.rawWrite(frm_header);
						}
						
						// Actual (un)compressed frame
						chk_header[0].size=6+fsize;
						chk_header[0].typ=frame_type; // FLI Copy (16) or Brun (15)
						
						outfile.rawWrite(chk_header);
						outfile.rawWrite(data);		
					}
				}
				catch (FileException e)
				{
				    writeln("Exception caught creating output file!");
				}
			}
		}
        catch (FileException e)
		{
		    writeln("Exception caught opening input file!");
		}
	}
	else
	{
		writeln("Could not find the file ",fn);
	}
}